#include <allegro.h>
#include <sys/farptr.h>
#include <go32.h>
#include <dpmi.h>
#include <string.h>
#include <pc.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>




typedef unsigned char       BYTE;
typedef unsigned short      WORD;
typedef unsigned long       DWORD;
typedef int BOOL;

#ifndef FALSE
#define FALSE   0
#endif

#ifndef TRUE
#define TRUE    !FALSE
#endif

#define PCI_ANY_ID  ((WORD)(~0))


typedef struct {
    WORD vender_id;
    WORD device_id;
    WORD sub_vender_id;
    WORD sub_device_id;
    WORD device_bus_number;
} PCI_DEV;


typedef struct {
    WORD vender_id;
    WORD device_id;
    WORD sub_vender_id;
    WORD sub_device_id;
    int  io_timing;
    int  dma_method;
    void (*ddma_setting_function)(WORD ddma_base, int dma_channel, int dma_channel2);
    char *string;
} DEVICE_LIST;

typedef struct {
    BYTE master;
    BYTE sb_pcm;
    BYTE opl;
    BYTE cd_audio;
    BYTE line_in;
    BYTE microphone;
    BYTE output_gain;
} SB_MIXER;

typedef struct {
    int wss_base;
    int sb_base;
    int irq;
    int dma;
    int mpu_base;
    int adlib_base;
    int game_base;
    int sb_ver;
    int dma_method;
    int io_timing;
    int ddma_base;
    int irq_trigger_mode;
}SB_RESOURCE;


#define DMA_AUTO                0
#define DMA_DDMA                1
#define DMA_LEGACY_DMA          2
#define DMA_WB_DMA              3
#define DMA_WRITE_SNOOP         4

#define IO_AUTO                 -1
#define IO_SUBTRACTIVE          0
#define IO_POSITIVE             1



/*
 *  grobal variables
 */

static int g_mcr_base = 0;
static int g_ddma_base = 0;
static int g_sb_base = 0;
static int g_wss_base = 0;



static BOOL find_known_device(int *io_timing, int *dma_method,
                              void (**ddma_setting_function)(WORD, int, int) );
void w_enter_critical(void);
void w_exit_critical(void);
static BOOL pci_read_config_byte(PCI_DEV *pci, int index, BYTE *data);
static BOOL pci_read_config_word(PCI_DEV *pci, int index, WORD *data);
static BOOL pci_read_config_dword(PCI_DEV *pci, int index, DWORD *data);
static BOOL pci_write_config_byte(PCI_DEV *pci, int index, BYTE data);
static BOOL pci_write_config_word(PCI_DEV *pci, int index, WORD data);
static BOOL pci_write_config_dword(PCI_DEV *pci, int index, DWORD data);
static BOOL set_pci_hardware_interrupt(PCI_DEV *pci, BYTE int_pin, BYTE irq);
static BOOL check_pci_bios(void);
static BOOL find_pci_device(PCI_DEV *pci);

static void logerror(const char *text,...)
{
    va_list arg;

    va_start(arg,text);
    vfprintf(stdout,text,arg);
    va_end(arg);
}

static void message(const char *text,...)
{
    va_list arg;

    va_start(arg,text);
    vfprintf(stdout,text,arg);
    va_end(arg);
}


/*
 *  chipset dependent DDMA setup funtions
 */

static void sis630_ddma(WORD ddma_base, int dma_channel, int dma_channel2)
{
    PCI_DEV pci;
    BYTE bdata;

    pci.vender_id     = 0x1039;             // SiS
    pci.device_id     = 0x0008;             // PCI-ISA Bridge
    pci.sub_vender_id = PCI_ANY_ID;
    pci.sub_device_id = PCI_ANY_ID;
    if(find_pci_device(&pci) == FALSE){
        pci.device_id     = 0x0018;         // PCI-ISA Bridge
        if(find_pci_device(&pci) == FALSE){
            message("sis630_ddma: No SiS PCI-ISA Bridge found\n");
            return;
        }
    }
    bdata = 0;
    if(dma_channel >= 0){
        bdata |= 1 << dma_channel;                          /* set DDMA channel */
    }
    if(dma_channel2 >= 0){
        bdata |= 1 << dma_channel2;                         /* set DDMA channel */
    }
    pci_write_config_byte(&pci, 0x49, bdata);
    bdata = (BYTE)((ddma_base & 0xF0) | 0x0001);        /* DDMA I/O base(lower 8bit) | DDMA enable flag*/
    pci_write_config_byte(&pci, 0x4A, bdata);
    bdata = (BYTE)(ddma_base >> 8);                     /* DDMA I/O base(higher 8bit) */
    pci_write_config_byte(&pci, 0x4B, bdata);
}

static void sis5595_ddma(WORD ddma_base, int dma_channel, int dma_channel2)
{
    PCI_DEV pci;
    BYTE bdata;

    pci.vender_id     = 0x1039;             // SiS
    pci.device_id     = 0x0008;             // PCI-ISA Bridge
    pci.sub_vender_id = PCI_ANY_ID;
    pci.sub_device_id = PCI_ANY_ID;
    if(find_pci_device(&pci) == FALSE){
        message("sis5595_ddma: No SiS PCI-ISA Bridge found\n");
        return;
    }
    bdata = 0;
    if(dma_channel >= 0){
        bdata |= 1 << dma_channel;                          /* set DDMA channel */
    }
    if(dma_channel2 >= 0){
        bdata |= 1 << dma_channel2;                         /* set DDMA channel */
    }

    pci_write_config_byte(&pci, 0x84, bdata);
    bdata = (BYTE)((ddma_base & 0xF0) | 0x0001);        /* DDMA I/O base(lower 8bit) | DDMA enable flag*/
    pci_write_config_byte(&pci, 0x80, bdata);
    bdata = (BYTE)(ddma_base >> 8);                     /* DDMA I/O base(higher 8bit) */
    pci_write_config_byte(&pci, 0x81, bdata);
}

static void intel_piix4_ddma(WORD ddma_base, int dma_channel, int dma_channel2)
{
/* dma_channel2 is ignored */

    PCI_DEV pci;
    WORD wdata;

    pci.vender_id     = 0x8086;             // Intel
    pci.device_id     = 0x7110;             // PIIX4
    pci.sub_vender_id = PCI_ANY_ID;
    pci.sub_device_id = PCI_ANY_ID;
    if(find_pci_device(&pci) == FALSE){
        message("intel_piix4_ddma: No PIIX4 found\n");
        return;
    }
    /*  set ddma enable bit */
    wdata = 0x02 << (dma_channel * 2);
    pci_write_config_word(&pci, 0x90, wdata);
    if(dma_channel >= 4){
        /*  DMA channel 4 to 7  */
        ddma_base += 0x40;
        pci_write_config_word(&pci, 0x94, ddma_base & 0xFFC0);
    }else{
        /*  DMA channel 0 to 3  */
        pci_write_config_word(&pci, 0x92, ddma_base & 0xFFC0);
    }
}

static void via_ddma(WORD ddma_base, int dma_channel, int dma_channel2)
{
/* dma_channel2 is ignored */

    PCI_DEV pci;
    int offset;
    WORD wdata;

    while(1){       // this is in place of goto-label
        pci.sub_vender_id = PCI_ANY_ID;
        pci.sub_device_id = PCI_ANY_ID;
        pci.vender_id     = 0x1106;             // VIA
        pci.device_id     = 0x0586;             // VT82C586
        if(find_pci_device(&pci) == TRUE) break;
        pci.device_id     = 0x0596;             // VT82C596
        if(find_pci_device(&pci) == TRUE) break;
        pci.device_id     = 0x0686;             // VT82C686
        if(find_pci_device(&pci) == TRUE) break;
        pci.vender_id     = 0x1022;             // AMD
        pci.device_id     = 0x7408;             // AMD-756
        if(find_pci_device(&pci) == TRUE) break;
        message("via_ddma: No VIA VT82C586/596/686/AMD-756 found\n");
        return;
    }
    dma_channel &= 0x07;
    if(dma_channel == 4) return;
    ddma_base += dma_channel * 0x10;
    offset = 0x60 + (dma_channel * 2);
    wdata  = (ddma_base & 0xFFF0) | 0x0008;
    pci_write_config_word(&pci, offset, wdata);
}

static DEVICE_LIST dev_list[] = {
    { 0x8086, 0x7110, PCI_ANY_ID, PCI_ANY_ID, IO_SUBTRACTIVE, DMA_DDMA,       intel_piix4_ddma, "Intel PIIX4" },
//    { 0x1106, 0x0586, PCI_ANY_ID, PCI_ANY_ID, IO_SUBTRACTIVE, DMA_LEGACY_DMA, NULL,             "VIA VT82C586" },
//    { 0x1106, 0x0596, PCI_ANY_ID, PCI_ANY_ID, IO_SUBTRACTIVE, DMA_LEGACY_DMA, NULL,             "VIA VT82C596" },
    { 0x1106, 0x0686, PCI_ANY_ID, PCI_ANY_ID, IO_SUBTRACTIVE, DMA_DDMA,         via_ddma,       "VIA VT82C686" },
//    { 0x1022, 0x7408, PCI_ANY_ID, PCI_ANY_ID, IO_SUBTRACTIVE, DMA_DDMA,       via_ddma,         "AMD-756" },
    { 0x1039, 0x5591, PCI_ANY_ID, PCI_ANY_ID, IO_SUBTRACTIVE, DMA_DDMA,       sis5595_ddma,     "SiS5591" },
    { 0x1039, 0x0530, PCI_ANY_ID, PCI_ANY_ID, IO_SUBTRACTIVE, DMA_DDMA,       sis5595_ddma,     "SiS530" },
    { 0x1039, 0x0620, PCI_ANY_ID, PCI_ANY_ID, IO_SUBTRACTIVE, DMA_DDMA,       sis5595_ddma,     "SiS620" },
//    { 0x1039, 0x0745, PCI_ANY_ID, PCI_ANY_ID, IO_SUBTRACTIVE, DMA_DDMA,       sis630_ddma,      "SiS745" },
    { 0x1039, 0x0735, PCI_ANY_ID, PCI_ANY_ID, IO_SUBTRACTIVE, DMA_DDMA,       sis630_ddma,      "SiS735" },
    { 0x1039, 0x0733, PCI_ANY_ID, PCI_ANY_ID, IO_SUBTRACTIVE, DMA_DDMA,       sis630_ddma,      "SiS733" },
    { 0x1039, 0x0730, PCI_ANY_ID, PCI_ANY_ID, IO_SUBTRACTIVE, DMA_DDMA,       sis630_ddma,      "SiS730" },
    { 0x1039, 0x0635, PCI_ANY_ID, PCI_ANY_ID, IO_SUBTRACTIVE, DMA_DDMA,       sis630_ddma,      "SiS635" },
    { 0x1039, 0x0633, PCI_ANY_ID, PCI_ANY_ID, IO_SUBTRACTIVE, DMA_DDMA,       sis630_ddma,      "SiS633" },
    { 0x1039, 0x0630, PCI_ANY_ID, PCI_ANY_ID, IO_SUBTRACTIVE, DMA_DDMA,       sis630_ddma,      "SiS630" },
    { 0x1039, 0x0540, PCI_ANY_ID, PCI_ANY_ID, IO_SUBTRACTIVE, DMA_DDMA,       sis630_ddma,      "SiS540" },
    { 0x1039, 0x0008, PCI_ANY_ID, PCI_ANY_ID, IO_SUBTRACTIVE, DMA_DDMA,       sis630_ddma,      "SiS0008 ISA Bridge" },

    /*  Add chipset information here  */

    { PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID, IO_SUBTRACTIVE, DMA_LEGACY_DMA, NULL, NULL }
};


static BOOL find_known_device(int *io_timing, int *dma_method,
                              void (**ddma_setting_function)(WORD, int, int) )
{
    int result = FALSE;
    int index;
    PCI_DEV pci;

    for(index = 0; TRUE ; index += 1){
        if(dev_list[index].vender_id == PCI_ANY_ID){
            *io_timing  = dev_list[index].io_timing;
            *dma_method = dev_list[index].dma_method;
            *ddma_setting_function = NULL;
            break;
        }
        pci.vender_id     = dev_list[index].vender_id;
        pci.device_id     = dev_list[index].device_id;
        pci.sub_vender_id = dev_list[index].sub_vender_id;
        pci.sub_device_id = dev_list[index].sub_device_id;
        if(find_pci_device(&pci) == TRUE){
            *io_timing  = dev_list[index].io_timing;
            *dma_method = dev_list[index].dma_method;
            *ddma_setting_function = dev_list[index].ddma_setting_function;
            result = TRUE;
            message("%s found.\n", dev_list[index].string);
            break;
        }
    }
    return result;
}



/*
 *  interrupt disable/enable funtion
 */

static DWORD tasmania = 0;

void w_enter_critical(void)
{
    __dpmi_regs r;

    tasmania += 1;
    if(tasmania == 1){
        r.x.ax = 0x1681;
        __dpmi_int(0x2F, &r);
        asm volatile("cli");
    }
}

void w_exit_critical(void)
{
    __dpmi_regs r;

    tasmania -= 1;
    if(tasmania == 0){
        r.x.ax = 0x1682;
        __dpmi_int(0x2F, &r);
        asm volatile("sti");
    }
}


/*
 *  PCI BIOS helper funtions
 */

static BOOL pci_read_config_byte(PCI_DEV *pci, int index, BYTE *data)
{
    __dpmi_regs r;
    BOOL result = TRUE;

    r.d.eax = 0x0000B108;                       /* read config byte */
    r.d.ebx = (DWORD)pci->device_bus_number;
    r.d.edi = (DWORD)index;
    __dpmi_int(0x1a, &r);
    if( r.h.ah != 0 ){
        logerror("pci read config byte failed\n");
        result = FALSE;
        r.d.ecx = 0;
    }
    *data = (BYTE)r.d.ecx;
    return result;
}

static BOOL pci_read_config_word(PCI_DEV *pci, int index, WORD *data)
{
    __dpmi_regs r;
    BOOL result = TRUE;

    r.d.eax = 0x0000B109;                       /* read config word */
    r.d.ebx = (DWORD)pci->device_bus_number;
    r.d.edi = (DWORD)index;
    __dpmi_int(0x1a, &r);
    if( r.h.ah != 0 ){
        logerror("pci read config word failed\n");
        result = FALSE;
        r.d.ecx = 0;
    }
    *data = (WORD)r.d.ecx;
    return result;
}

static BOOL pci_read_config_dword(PCI_DEV *pci, int index, DWORD *data)
{
    __dpmi_regs r;
    BOOL result = TRUE;

    r.d.eax = 0x0000B10A;                       /* read config dword */
    r.d.ebx = (DWORD)pci->device_bus_number;
    r.d.edi = (DWORD)index;
    __dpmi_int(0x1a, &r);
    if( r.h.ah != 0 ){
        logerror("pci read config dword failed\n");
        result = FALSE;
        r.d.ecx = 0;
    }
    *data = (DWORD)r.d.ecx;
    return result;
}

static BOOL pci_write_config_byte(PCI_DEV *pci, int index, BYTE data)
{
    __dpmi_regs r;
    BOOL result = TRUE;

    r.d.eax = 0x0000B10B;                       /* write config byte */
    r.d.ebx = (DWORD)pci->device_bus_number;
    r.d.ecx = (DWORD)data;
    r.d.edi = (DWORD)index;
    __dpmi_int(0x1a, &r);
    if( r.h.ah != 0 ){
        logerror("pci write config byte failed\n");
        result = FALSE;
    }
    return result;
}

static BOOL pci_write_config_word(PCI_DEV *pci, int index, WORD data)
{
    __dpmi_regs r;
    BOOL result = TRUE;

    r.d.eax = 0x0000B10C;                       /* write config word */
    r.d.ebx = (DWORD)pci->device_bus_number;
    r.d.ecx = (DWORD)data;
    r.d.edi = (DWORD)index;
    __dpmi_int(0x1a, &r);
    if( r.h.ah != 0 ){
        logerror("pci write config word failed\n");
        result = FALSE;
    }
    return result;
}

static BOOL pci_write_config_dword(PCI_DEV *pci, int index, DWORD data)
{
    __dpmi_regs r;
    BOOL result = TRUE;

    r.d.eax = 0x0000B10D;                       /* write config dword */
    r.d.ebx = (DWORD)pci->device_bus_number;
    r.d.ecx = (DWORD)data;
    r.d.edi = (DWORD)index;
    __dpmi_int(0x1a, &r);
    if( r.h.ah != 0 ){
        logerror("pci write config dword failed\n");
        result = FALSE;
    }
    return result;
}


static BOOL set_pci_hardware_interrupt(PCI_DEV *pci, BYTE int_pin, BYTE irq)
{
    __dpmi_regs r;
    BOOL result = TRUE;

    r.d.eax = 0x0000B10F;                       /* set pci hardware interrupt */
    r.d.ebx = (DWORD)pci->device_bus_number;
    r.h.cl  = int_pin;
    r.h.ch  = irq;
    r.x.ds  = 0xF000;
    __dpmi_int(0x1a, &r);
    if( r.h.ah != 0 ){
        logerror("set pci hardware interrupt failed.\n");
        result = FALSE;
    }
    pci_write_config_byte(pci, 0x3C, irq);
    return result;
}



static BOOL check_pci_bios(void)
{
    __dpmi_regs r;
    BOOL result = TRUE;

    r.d.eax = 0x0000B101;                   // PCI BIOS - INSTALLATION CHECK
    r.d.edi = 0x00000000;
    __dpmi_int(0x1a, &r);
    if( r.d.edx != 0x20494350 ){            // ' ICP'
        result = FALSE;
    }
    return result;
}


static BOOL find_pci_device(PCI_DEV *pci)
{
    __dpmi_regs r;
    WORD wdata;

    r.d.eax = 0x0000B102;                   // PCI BIOS - FIND PCI DEVICE
    r.d.ecx = pci->device_id;               // device ID
    r.d.edx = pci->vender_id;               // vendor ID
    r.d.esi = 0x00000000;                   // device index
    __dpmi_int(0x1a, &r);
    if( r.h.ah != 0 ){
        return FALSE;                       // no specified device found
    }
    pci->device_bus_number = r.x.bx;        // save the device & bus number
    if(pci->sub_vender_id != PCI_ANY_ID){
        /* get subsystem vender id */
        if(pci_read_config_word(pci, 0x2C, &wdata) == FALSE) return FALSE;
        if(wdata != pci->sub_vender_id) return FALSE;
    }
    if(pci->sub_device_id != PCI_ANY_ID){
        /* get subsystem device id */
        if(pci_read_config_word(pci, 0x2E, &wdata) == FALSE) return FALSE;
        if(wdata != pci->sub_device_id) return FALSE;
    }
    /* device found */
    return TRUE;
}


static void enable_pci_io_access(PCI_DEV *pci)
{
    WORD wdata;

    pci_read_config_word(pci, 0x04, &wdata);
    /*  set I/O access enable bit  */
    wdata |= 0x0001;
    pci_write_config_word(pci, 0x04, wdata);
}

#if 0

static void enable_pci_memory_access(PCI_DEV *pci)
{
    WORD wdata;

    pci_read_config_word(pci, 0x04, &wdata);
    /*  set memory access enable bit  */
    wdata |= 0x0002;
    pci_write_config_word(pci, 0x04, wdata);
}

#endif

static void enable_pci_bus_master(PCI_DEV *pci)
{
    WORD wdata;

    pci_read_config_word(pci, 0x04, &wdata);
    /*  set bus master enable bit  */
    wdata |= 0x0004;
    pci_write_config_word(pci, 0x04, wdata);
}



/*
 *  SB functions
 */


static void sb_write_mixer(BYTE index, BYTE data)
{
    outportb(g_sb_base + 4, index);
    outportb(g_sb_base + 5, data);
}

#if 0

static BYTE sb_read_mixer(BYTE index)
{
    outportb(g_sb_base + 4, index);
    return inportb(g_sb_base + 5);
}

static void sb_write_ctrlreg(BYTE index, BYTE data)
{
    outportb(g_sb_base + 4, index | 0xC0);
    outportb(g_sb_base + 5, data);
}

static BYTE sb_read_ctrlreg(BYTE index)
{
    outportb(g_sb_base + 4, index | 0xC0);
    return inportb(g_sb_base + 5);
}
#endif


static int sb_read_dsp(void)
{
   int x;

   for (x=0; x<0xFFFF; x++)
      if (inportb(0x0E + g_sb_base) & 0x80)
     return inportb(0x0A + g_sb_base);

   return -1;
}

static int sb_reset_dsp(int d0)
{
    int val;
    int cwait;
    int result;
    int d7;

    result = -1;
    cwait = 8;
    while(cwait <= 1024){
        w_enter_critical();
        outportb(0x06 + g_sb_base, d0);
        d7 = 0;
        while(d7 < cwait){
            inportb(0x06 + g_sb_base);
            d7 += 1;
        }
        outportb(0x06 + g_sb_base, 0);
        w_exit_critical();
        for(d7 = 0 ; d7 < 4 ; d7 += 1){
            val = sb_read_dsp();
            if(val == -1) continue;
            if(val == 0xAA){
                result = 0;
                break;
            }
            break;
        }
        if(result == 0) break;
        cwait = cwait * 2;
    }
    return result;
}




#if 0

static int get_irq_edge_or_level(int irq)
{
    int status;

    /*  I/O 04D0h
     *      bit 7   irq 7   0=Edge Triggered mode; 1=Level Triggered mode
     *      bit 6   irq 6   0=Edge Triggered mode; 1=Level Triggered mode
     *                          .
     *                          .
     *                          .
     *      bit 0   irq 0   0=Edge Triggered mode; 1=Level Triggered mode
     *
     *  I/O 04D1h
     *      bit 7   irq 15  0=Edge Triggered mode; 1=Level Triggered mode
     *      bit 6   irq 14  0=Edge Triggered mode; 1=Level Triggered mode
     *                          .
     *                          .
     *                          .
     *      bit 0   irq 8   0=Edge Triggered mode; 1=Level Triggered mode
     *
     *
     *  return == 0     edge  triggered
     *         != 0     level triggered
     */

    if(irq > 15) return 0;
    if(irq < 8){
        status = inportb(0x04D0) & (1 << irq);
    }else{
        irq -= 8;
        status = inportb(0x04D1) & (1 << irq);
    }
    if(status != 0) status = 1;

    return status;
}

#endif

static void set_irq_edge_or_level(int irq, int edge_or_level)
{
    /*
     *  edge_or_level == 0     set edge  triggered
     *                != 0     set level triggered
     */

    int bitmap;

    if(irq > 15) return;
    if(edge_or_level != 0) edge_or_level = 1;
    if(irq < 8){
        bitmap = inportb(0x04D0);
        bitmap = bitmap & ~(1 << irq);
        bitmap = bitmap | (edge_or_level << irq);
        outportb(0x04D0, bitmap);
    }else{
        irq -= 8;
        bitmap = inportb(0x04D1);
        bitmap = bitmap & ~(1 << irq);
        bitmap = bitmap | (edge_or_level << irq);
        outportb(0x04D1, bitmap);
    }
}


static void get_sb_resource(SB_RESOURCE *res)
{
    /*  variables  */
    res->irq              = get_config_int("resource", "IRQ", 7) & 0x0F;
    res->irq_trigger_mode = get_config_int("resource", "IRQ_TRIGGER", 1) & 0x01;

    /*  we use the fixed parameters.    */
    res->wss_base    = get_config_int("resource", "WSS",      0x534) & ~0x0003;
    res->sb_base     = get_config_int("resource", "SB16",     0x220) & ~0x000F;
    res->dma         = get_config_int("resource", "DMA",          1) & 0x03;
    res->mpu_base    = get_config_int("resource", "MPU401",   0x330) & ~0x0007;
    res->adlib_base  = get_config_int("resource", "ADLIB",    0x388) & ~0x0007;
    res->game_base   = get_config_int("resource", "GAMEPORT", 0x200) & ~0x0007;
    res->sb_ver           = get_config_int("resource", "DSP_VER",  0x302);
//    res->dma_method  = get_config_int("resource", "DMA_METHOD", 0);
//    res->io_timing   = get_config_int("resource", "IO_TIMING",  IO_AUTO);
//    res->ddma_base   = get_config_int("resource", "DDMA_BASE",  0x1000) & ~0x000F;
}


static void print_sb_resource(SB_RESOURCE *res)
{
    char irq[2][8] = { "Edge", "Level" };

    printf("[Resource]\n");
    printf("  WSS: %3Xh, ", res->wss_base);
    printf("SB: %3Xh, ", res->sb_base);
    printf("IRQ:%2d, ", res->irq);
    printf("DMA: %1X\n", res->dma);
    printf("  MPU401: %3Xh, ", res->mpu_base);
    printf("ADLIB: %3Xh, ", res->adlib_base);
    printf("GAMEPORT: %3Xh\n", res->game_base);
    printf("  IRQ %d is %s triggered.\n", res->irq, irq[res->irq_trigger_mode & 0x01]);
}


static void get_mixer_level(SB_MIXER *mixer)
{
    mixer->master      = (BYTE)get_config_int("volume", "MASTER", 31);
    mixer->sb_pcm      = (BYTE)get_config_int("volume", "PCM", 27);
    mixer->opl         = (BYTE)get_config_int("volume", "OPL", 27);
    mixer->cd_audio    = (BYTE)get_config_int("volume", "CD", 0);
    mixer->line_in     = (BYTE)get_config_int("volume", "LINE", 0);
    mixer->microphone  = (BYTE)get_config_int("volume", "MIC", 0);
    mixer->output_gain = (BYTE)get_config_int("volume", "OUTPUT_GAIN", 2);
}

static void print_mixer_level(SB_MIXER *mixer)
{
    printf("[Volume]\n");
    printf("  MASTER:%2d, ", mixer->master);
    printf("PCM:%2d, ", mixer->sb_pcm);
    printf("OPL:%2d, ", mixer->opl);
    printf("CD:%2d, ", mixer->cd_audio);
    printf("LINE:%2d, ", mixer->line_in);
    printf("MIC:%2d\n", mixer->microphone);
}



static int get_irq(PCI_DEV *pci)
{
    BYTE irq;

    /* get irq number */
    pci_read_config_byte(pci, 0x3C, &irq);
    return (int)irq;
}



static void copy_to_dos_memory(int selector, BYTE *ptr, int length)
{
    int d0;

    _farsetsel(selector);
    d0 = 0;
    while(d0 < length){
        _farnspokeb(d0, *(ptr + d0));
        d0 += 1;
    }
}


static BOOL _INT2E(char* commandline)
{
    __dpmi_regs r;
    int segment;
    int selector;
    int d0;
    int d1;
    char* a0;

    segment = __dpmi_allocate_dos_memory(1024, &selector);
    if(segment == -1) return FALSE;

    a0 = (char*)malloc(512);
    if(a0 == NULL) goto _INT2E_error;
    d0 = strlen(commandline);
    if(d0 > 255) goto _INT2E_error;

    d1 = 0;
    while(d1 < d0){
        a0[d1+1] = commandline[d1];
        d1 += 1;
    }
    a0[d1+1] = 0x0D;
    a0[d1+2] = 0;
    a0[0] = d0;

    copy_to_dos_memory(selector, a0, d0 + 2);

    r.x.ds = segment;
    r.x.si = 0;
    __dpmi_int(0x2E, &r);
    if( r.x.ax == 0xFFFF ) goto _INT2E_error;

    __dpmi_free_dos_memory(selector);
    return TRUE;

_INT2E_error:
    __dpmi_free_dos_memory(selector);
    return FALSE;
}

static void set_blaster_environment(int port, int irq, int dma, int mpu401, int dspver)
{
    char sb[]   = "SET BLASTER=A%x I%d D%d T%d";
    char sb16[] = "SET BLASTER=A%x I%d D%d H%d T%d P%x";
    char a0[512];
    int type = 0;

    if( (dspver >> 8) >= 4 ){
        sprintf(a0, sb16, port, irq, dma, dma, 6, mpu401);
    }else{
        if( (dspver >> 8) == 3 ){
            type = 4;
        }else if( (dspver >> 8) == 2 ){
            type = 3;
        }else if( (dspver >> 8) == 1 ){
            type = 1;
        }
        sprintf(a0, sb, port, irq, dma, type);
    }
    message("\n%s\n", &a0[4]);
    _INT2E(a0);
}

static void set_wsscfg_environment(int port, int irq, int dma)
{
    char wss[]   = "SET WSSCFG=A%x I%d D%d";
    char a0[512];

    sprintf(a0, wss, port, irq, dma);
//    message("\n%s\n", &a0[4]);
    _INT2E(a0);
}


static void write_mcreg(int index, int value)
{
    outportb(g_mcr_base + 0x0E, index);
    outportb(g_mcr_base + 0x0F, value);
}

static void setup_mcregs(void)
{
    write_mcreg(0x01, 0x05);
    write_mcreg(0x02, 0x07);
    write_mcreg(0x03, 0x2E);
    write_mcreg(0x04, 0xD6);
    write_mcreg(0x05, 0x3D);
    write_mcreg(0x06, 0xC3);
    write_mcreg(0x0A, 0x01);
    write_mcreg(0x0B, 0x03);
    write_mcreg(0x0C, 0xDE);
    write_mcreg(0x0D, 0x01);
}

static void wssout(BYTE index, BYTE data)
{
    outportb(g_wss_base + 4, index);
    outportb(g_wss_base + 5, data);
}

#if 0
static BYTE wssin(BYTE d0)
{
    outportb(g_wss_base + 4, d0);
    return inportb(g_wss_base + 5);
}
#endif


static void setup_wss_volumes(SB_MIXER mixer)
{
    BYTE level;
    BYTE mute = 0x80;

    sb_write_mixer(0x00, 0x00);                 //  reset
    sb_write_mixer(0x0E, 0x31);                 //  mono
    sb_write_mixer(0x02, 0xFF);                 //  master volume
    sb_write_mixer(0x04, 0xFF);                 //  sb-pcm volume
    sb_write_mixer(0x06, 0xFF);                 //  opl volume

    wssout(10, 0x02);           // interrupt enable
    wssout(12, 0xE0);           // extened mode

    wssout(0, 0);               //  mixout left
    wssout(1, 0);               //  mixout right

    /*  master  */
    level = (mixer.master & 0x1F) << 1;         //  0:0dB, 31:-93dB  (3.0dB step)
    if(level == 0x3E) level |= mute;            //  if min volume, mute enable
    wssout(22, level);                          //  master volume left
    wssout(23, level);                          //  master volume right

    /*  dac  */
    level = (mixer.sb_pcm & 0x1F);              //  0:0dB, 31:-46.5dB  (1.5dB step)
    if(level == 0x1F) level |= mute;            //  if min volume, mute enable
    wssout( 6, level);                          //  dac volume left
    wssout( 7, level);                          //  dac volume right

    /*  opl  */
    level = (mixer.opl & 0x0F) << 1;            //  0:12dB, 4:0dB, 15:-33dB  (3.0dB step)
    if(level == 0x1E) level |= mute;            //  if min volume, mute enable
    wssout( 4, level);                          //  opl volume left
    wssout( 5, level);                          //  opl volume right

    /*  cd audio  */
    level = (mixer.cd_audio & 0x0F) << 1;       //  0:12dB, 4:0dB, 15:-33dB  (3.0dB step)
    if(level == 0x1E) level |= mute;            //  if min volume, mute enable
    wssout( 2, level);                          //  cd audio volume left
    wssout( 3, level);                          //  cd audio volume right

    /*  aux  (fixed level) */
    level = (0x0F & 0x0F) << 1;                 //  0:12dB, 4:0dB, 15:-33dB  (3.0dB step)
    if(level == 0x1E) level |= mute;            //  if min volume, mute enable
    wssout(16, level);                          //  aux volume left
    wssout(17, level);                          //  aux volume right

    /*  line in  */
    level = (mixer.line_in & 0x0F) << 1;        //  0:12dB, 4:0dB, 15:-33dB  (3.0dB step)
    if(level == 0x1E) level |= mute;            //  if min volume, mute enable
    wssout(18, level);                          //  line in audio volume left
    wssout(19, level);                          //  line in audio volume right

    /*  mic  */
    level = (mixer.microphone & 0x0F) << 1;     //  0:12dB, 4:0dB, 15:-33dB  (3.0dB step)
    if(level == 0x1E) level |= mute;            //  if min volume, mute enable
    wssout(20, level);                          //  mic audio volume left
    wssout(21, level);                          //  mic audio volume right
}




int main(int argc, char* argv[])
{
    PCI_DEV pci;
    SB_MIXER mixer;
    SB_RESOURCE resource;
    int ev1935_prev_irq;
    int io_timing;
    int dma_method;
    DWORD d0;
    WORD d1;
    void (*ddma_setting_function)(WORD, int, int);

//    printf("ALS4000 DOS Setup Utility  2002.7.23  by saka <als4k@anet.ne.jp> \n\n");

    /*  get resource & volume setting from ev1935.cfg  */
    allegro_init();
    set_config_file("ev1935.cfg");
    get_sb_resource(&resource);
    get_mixer_level(&mixer);

    if(check_pci_bios() == FALSE){
        printf("No PCI BIOS found\n");
        return FALSE;
    }

    /*  EV1935 detection */
    pci.vender_id = 0x1045;         // OPTi
    pci.device_id = 0xC935;         // 935
    pci.sub_vender_id = PCI_ANY_ID;
    pci.sub_device_id = PCI_ANY_ID;
    if(find_pci_device(&pci) == FALSE){
        printf("No EV1935 found\n");
        return FALSE;
    }
    printf("EV1935 found.\n");

    /*  compare chipsets list with current one  */
    if(find_known_device(&io_timing, &dma_method, &ddma_setting_function) == FALSE){
        printf("No DDMA support chipset found\n");
        return FALSE;
    }
    if(dma_method != DMA_DDMA){
        printf("No DDMA support chipset found\n");
        return FALSE;
    }

    /*  I/O access & bus master enable  */
    enable_pci_io_access(&pci);
    enable_pci_bus_master(&pci);

    /*  get ev1935's current irq  */
    ev1935_prev_irq = get_irq(&pci);
    /*  if irq is specified,  */
    if(resource.irq > 0){
        /*  set specified irq */
        set_pci_hardware_interrupt(&pci, 0x0A, resource.irq);
    }
    /*  get irq number  */
    resource.irq = get_irq(&pci);

    /*  setup irq trigger mode  */
    set_irq_edge_or_level(resource.irq, resource.irq_trigger_mode);


    /*  set legacy base I/O address registers  */
    pci_write_config_word(&pci, 0x44, resource.sb_base);        // SB base
    pci_write_config_word(&pci, 0x46, resource.wss_base);       // WSS base
    pci_write_config_word(&pci, 0x48, resource.adlib_base);     // AdLib base
    pci_write_config_word(&pci, 0x4A, resource.mpu_base);       // MPU401 base
    pci_write_config_word(&pci, 0x4C, resource.game_base);      // GAMEPORT base

    /*  get base registers for later use  */
    pci_read_config_dword(&pci, 0x10, &d0);
    g_mcr_base  = (int)(d0 & ~0x0F);

    pci_read_config_dword(&pci, 0x14, &d0);
    g_ddma_base = (int)(d0 & ~0x7F);

    pci_read_config_word(&pci, 0x46, &d1);
    g_wss_base  = (int)(d1 & ~0x03);
    g_wss_base  -= 4;

    pci_read_config_word(&pci, 0x44, &d1);
    g_sb_base   = (int)(d1 & ~0x03);


    /*  legecy I/O enable?  */
    pci_write_config_byte(&pci, 0x61, 0x96);             // magic number?
    pci_write_config_byte(&pci, 0x62, 0xC0);             // magic number?

    /*  others  */
    pci_write_config_dword(&pci, 0x54, 0x00000000);      // IRQ Driveback Protocol Address Register? disable?

    /*  write the password to unlock mcregs  */
    w_enter_critical();
    outportb(0x0F8D, 0xE4);
    outportb(0x0F8D, 0x80);
    w_exit_critical();

    /*  setup mcregs  */
    setup_mcregs();

    /*  setup volumes, and enable interrupt */
    setup_wss_volumes(mixer);

    /*  call chipset dependent DDMA setup code  */
    if(ddma_setting_function != NULL){
        (*ddma_setting_function)(g_ddma_base, 1, 0);
    }

    /*  print ev1935 infomations  */
    printf("MCREG: %04Xh, DDMA: %04Xh, IRQ: ", g_mcr_base, g_ddma_base);
    if(resource.irq == ev1935_prev_irq){
        printf("%d\n", resource.irq);
    }else{
        printf("%d -> %d\n", ev1935_prev_irq, resource.irq);
    }

    if(sb_reset_dsp(1) != 0){
        message("### SB reset failed ###\n");
    }

    /*  print current settings  */
    print_mixer_level(&mixer);
    print_sb_resource(&resource);

    /*  set BLASTER environment  */
    set_blaster_environment(resource.sb_base, resource.irq, resource.dma,
                            resource.mpu_base, resource.sb_ver);
    /*  set WSSCFG environment  */
    set_wsscfg_environment(resource.wss_base - 4, resource.irq, resource.dma);

    return 0;
}
